
// TBD: consider changing CThread constructor to accept an object
// derived from CAtom, which implements the Run() method, and then,
// when the user calls the thread.Start() method, it, in turn, calls
// the object's Run() method. Problem: I lose the convenience of the
// builtin inPort.

//*********************************************************************
// thread.h
//*********************************************************************
//
// By adding three addiional functions to the CThread class, the CThread
// class would look like the standard CApplication model.
//
// bool InitInstance();
// void TermInstance();
// bool onEvent(RPCEvent _event);
//
//*********************************************************************

#ifndef THREAD_H
#define THREAD_H

#ifndef FARBOOK_H
#include "atom.h"
#include "list.h"
#endif

//=====================================================================
// mutual exclusion protection
//
_FBDEF(CSemaphore)
class FBC_USE CSemaphore: public CAtom {
   public:
      FB_USE CSemaphore(short _start = 0, short _size = 1);
      virtual FB_USE ~CSemaphore();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      virtual FB_USE_(ulong) Wait(void);
      virtual FB_USE_(ulong) Release(void);

   protected:
#if   defined(FV_MSWIN16)
   #warning Semaphore not supported

#elif defined(FV_MSWIN32)
      HANDLE m_hSema4;

#elif defined(FV_MACINTOSH)
   #warning FV_MACINTOSH TBD

#elif defined(FV_UNIX)
   #warning FV_UNIX TBD

#endif // FV_UNIX
   }; // CSemaphore

//=====================================================================
// supplier vs consumer
//
// To use the CPort object as an event queue for a thread, derive an
// acceptable class from CItem to include members for the event code,
// any required parameters, and a response state. To use the event
// object, the client instantiates the object, assigning a agreeable
// event code, provides the required parameters, and sets the response
// state member to an initial value. Usually the possible states are
// some elaboration the enum {ready, running, done}. The client can
// then stick the event into the server's inport with the following
// code:
//
//    theServer->GetPort()->Post(myEvent);
//
// The client can determine the state of the service by examining the
// value of the response member, as in the following code:
//
//    while (myEvent->response != done);
//    ...continue;
//
// You can also add a CSemaphore member to the event, and Wait() on
// that.
//
// Of course, no matter how you choose to implement server=to-client
// communication, the server will have to cooperate. The CPort and the
// CThread class are kept simple to maximize your flexibility.
//
_FBDEF(CPort)
class FBC_USE CPort: public CAtom {
   public:
      FB_USE CPort(short _size);
      virtual FB_USE ~CPort();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      FB_USE_(bool)   Empty  (void);
      FB_USE_(ulong)  Number (void) { return list->Number(); }

      FB_USE_(void)   Post   (RPCItem _item);
      FB_USE_(PCItem) Receive(void);

   protected:
      PCSemaphore s1;
      PCSemaphore s2;
      PCSemaphore s3;
      PCList      list;
   }; // CPort

//=====================================================================
// restrict access to code
//
_FBDEF(CCriticalSection)
class FBC_USE CCriticalSection: public CAtom {
   public:
      FB_USE CCriticalSection(void);
      virtual FB_USE ~CCriticalSection();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      virtual FB_USE_(void) Enter(void);
      virtual FB_USE_(void) Leave(void);

   protected:
   
#if   defined(FV_MSWIN16)
   #warning Critical Section not supported

#elif defined(FV_MSWIN32)
      CRITICAL_SECTION m_cs;

#elif defined(FV_MACINTOSH)
   #warning FV_MACINTOSH TBD

#elif defined(FV_UNIX)
   #warning FV_UNIX TBD

#endif // FV_UNIX
   }; // CCriticalSection

//=====================================================================
// Class to simplify use of CCriticalSection. Create a smart pointer to
// a CLock variable at the top of a routine. When the routine quits, 
// the destructor will automatically release the critical section.
//
// function myFunction() {
//    PCLock lock = new CLock(myCriticalSection);
//    ...do critical stuff...
//    } // myFunction
//
// out-a-scope out-a-mind
//
_FBDEF(CLock)
class FBC_USE CLock: public CAtom {
   public:
      FB_USE CLock(RPCCriticalSection _section):
         section(_section) {
         section->Enter();
         }
      virtual FB_USE ~CLock() {
         section->Leave();
         }

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

   protected:
      PCCriticalSection section;
   }; // CLock

//=====================================================================
// Independent operation - separate stack, common data
//
// After constructing a CThread object, call the Start() method to
// begin the thread. You must override the Run() method to implement
// the thread. The thread must Stop() itself, noone else!
//
// Also, to operate in the farBook environment, your thread class name
// must begin with a capital-C, as in CMyThread, and you must define a
// ulong constant whose name must begin with a lower-case-i, as in
// iMyThread. This strangeness is because of the _FBDEF and RTTI_CPP
// macros, which implement farVIEW's run-time type checking. Actually,
// these rules are true for any class you derive within the farBook
// environment.
//
// Example Code:
//
// Derive a class from CThread:
//
//    const ulong iMyThread = 1000;
//
//    _FBDEF(CMyThread)
//    class FBC_USE CMyThread: public CThread {
//       public:
//          FB_USE CMyThread(void);
//          virtual FB_USE ~CMyThread();
//
//          virtual FB_USE_(short) Ima(void);
//          virtual FB_USE_(Pchar) IAm(void);
//          virtual FB_USE_(bool)  Isa(const short _type);
//
//          FB_USE_(void) Run(void);
//
//          bool Quit;    // in this example, when the parent sets this
//                       // variable to true, CMyThread::Run will stop.
//       }; // CMyThread
//
//---------------------------------------------------------------------
//
//    CMyThread::CMyThread(void):
//       CThread(),
//       Quit   (false) {
//       } // CMyThread
//
//---------------------------------------------------------------------
//
//    CMyThread::~CMyThread() {
//       } // ~CMyThread
//
//---------------------------------------------------------------------
// Macro to generate code for Ima(), IAm(), and Isa()
//
//    RTTI_CPP(MyThread, Thread)
//
//---------------------------------------------------------------------
//
//    void CMyThread::Run(void) {
//       while (!Quit) {     // when parent sets Quit, the thread stops
//          // Insert your control logic for the thread here
//          Pause(wait);
//          }
//       Stop();
//       } // Run
//
//---------------------------------------------------------------------
// The parent instantiates the thread using the Start() method as in
// the following two lines of code:
//
//    PCMyThread myThread = new CMyThread();
//    myThread->Start(32000);         // 32000 specifies the stack size
//
// The P-prefix on the class name (as in PCMyThread) is defined by the
// _FBDEF macro and declares a smart pointer, which will take care of
// deleting and destructing the object. Do not delete the object
// yourself. When it goes out of scope, it will decrement its use-count,
// and when that reaches zero, the object will self-destruct. As a
// general rule, you should always use smart pointers when using the
// classes provided in the farBook DLL.
//
// In the CMyThread example, when the parent wants the thread to quit,
// it simply uses the following line:
//
//    myThread->Quit = true;
//
// Note that the thread inherits an input port that other threads can
// use to post messages (derived from CItem) to the thread. The client
// posts a message with the following line:
//
//    myThread->GetPort()->Post(myMessage);
//
// The thread receives the message with the following line:
//
//    PCMessage message = GetPort()->Receive();
//
// By including a status flag member in the message that the server
// thread sets according to the state of the request, the client thread
// can keep track of progress. Just be sure that only the server thread
// sets the value of the flag member, and the client thread only
// examines the value. 'Two' many cooks will spoil the broth.
//
// Note: farSlang modules must never instantiate a farSlang thread
// object using the CThread class. A different class (C???) is
// available for the purpose.
//
enum THREADSTATE {THREAD_READY, THREAD_RUNNING, THREAD_ENDING, THREAD_STOPPED};

_FBDEF(CThread)
class FBC_USE CThread: public CItem {
   public:
      FB_USE CThread(short _size = 8);
      virtual FB_USE ~CThread();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

#ifndef _SLANGER
      FB_USE_(ulong) Start(ulong _sizeStack = 0);
      FB_USE_(void)  Stop (void);
      FB_USE_(void)  Join (void);
      virtual
      FB_USE_(void)  Run  (void);
      FB_USE_(void)  Log  (char _flag, Pchar _format, ...);
#endif // _SLANGER

      FB_USE_(PCPort) GetPort(void) { return port; }

      FB_USE_(ulong)  ThreadID(void) { return id; }

      FB_USE_(THREADSTATE) GetState(void) { return state; }

   protected:
	   static void   run(void *_arg);

      static ushort threads;

      ulong       id;
      PCPort      port;
      THREADSTATE state;
//?   ulong  then;
  }; // CThread

#endif // THREAD_H

